
/*
 * Routines to maintain a list of online users.
 *
 * Blitzed Services copyright (c) 2000-2002 Blitzed Services team
 *     E-mail: services@lists.blitzed.org
 * Based on ircservices-4.4.8:
 * copyright (c) 1996-1999 Andrew Church.
 *     E-mail: <achurch@dragonfire.net>
 * copyright (c) 1999-2000 Andrew Kempe.
 *     E-mail: <theshadow@shadowfire.org>
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * $Id: users.c,v 1.12 2003/11/05 00:33:42 stu Exp $
 */

#include "stdinc.h"
#include "tools.h"
#include "user.h"
#include "services.h"
#include "hash.h"
#include "language.h"
#include "memory.h"
#include "mystring.h"
#include "smysql.h"
#include "log.h"
#include "send.h"
#include "nickserv.h"
#include "chanserv.h"
#include "channel.h"
#include "akill.h"
#include "sessions.h"
#include "server.h"
#include "news.h"
#include "misc.h"
#include "memoserv.h"
#include "cache.h"

u_int32_t usercnt = 0, opcnt = 0, maxusercnt = 0;
time_t maxusertime;

int match_debug = 0;

static void mysql_quit_user(struct User *, const time_t, const char *);
static void display_entry_message(struct User *, unsigned int);

dlink_list user_list = { 0, 0, 0 };

/* Allocate a new struct User structure, fill in basic values, link it to the
 * overall list, and return it.  Always successful.
 */

static struct User *
new_user(const char *nick)
{
	unsigned int fields, rows;
	MYSQL_RES *result;
	struct User *user;

	user = scalloc(sizeof(struct User), 1);

	if (!nick)
		nick = "";
	
	strscpy(user->nick, nick, NICKMAX);

    hash_add_user(user);
    dlinkAdd(user, &user->unode, &user_list);
	
	user->real_id = mysql_findnick(nick);
	
	if (user->real_id)
		user->nick_id = mysql_getlink(user->real_id);
	else
		user->nick_id = 0;
	
	usercnt++;

	if (usercnt > maxusercnt) 
    {
		maxusercnt = usercnt;
		maxusertime = time(NULL);

		result = smysql_bulk_query(mysqlconn, &fields, &rows,
		    "REPLACE INTO stat (name, value) "
		    "VALUES ('maxusercnt', '%d')", maxusercnt);
		mysql_free_result(result);

		result = smysql_bulk_query(mysqlconn, &fields, &rows,
		    "REPLACE INTO stat (name, value) "
		    "VALUES ('maxusertime', '%lu')", maxusertime);
		mysql_free_result(result);
		
		if (LogMaxUsers) 
        {
			slog("user: New maximum user count: %d", maxusercnt);
			wallops(s_NickServ,
			    "New maximum user count: \2%d\2", maxusercnt);
		}
	}
	return user;
}

/* Change the nickname of a user, and move pointers as necessary. */

static void 
change_user_nick(struct User *user, const char *nick, int ni_changed)
{
	int status;

    if(user->nick[0])
        hash_del_client(user);

    strscpy(user->nick, nick, NICKMAX);
    hash_add_user(user);
	
    user->real_id = mysql_findnick(nick);
	
	if (user->real_id) 
    {
		if (!ni_changed) 
        {
			status = get_nick_status(user->nick_id);
			set_nick_status(user->real_id, status);
		}
		user->nick_id = mysql_getlink(user->real_id);
	} 
    else 
		user->nick_id = 0;
}

/* Remove and free a struct User structure. */

static void 
delete_user(struct User *user)
{
    struct Channel *c;
    dlink_node *ptr, *n_ptr;
    
	if (debug >= 2)
		slog("debug: delete_user() called");

	usercnt--;
	if (IsOper(user))
		opcnt--;

	cancel_user(user);
	
	if (debug >= 2)
		slog("debug: delete_user(): free user data");
	
	if (debug >= 2)
		slog("debug: delete_user(): remove from channels");
	
    DLINK_FOREACH_SAFE(ptr, n_ptr, user->chans.head)
    {
        c = ptr->data;
		chan_deluser(user, c);
		free(c);
	}
	
	if (debug >= 2)
		slog("debug: delete_user(): free founder data");
	
    DLINK_FOREACH_SAFE(ptr, n_ptr, user->founder_chans.head)
    {
        c = ptr->data;
		free(c);
	}
	
	if (debug >= 2)
		slog("debug: delete_user(): delete from list");

    hash_del_client(user);
    dlinkDelete(&user->unode, &user_list);

	if (debug >= 2)
		slog("debug: delete_user(): free user structure");
	
	free(user);
	
	if (debug >= 2)
		slog("debug: delete_user() done");
}

/* Return statistics.  Pointers are assumed to be valid. */

void 
get_user_stats(long *nusers, long *memuse)
{
	long count, mem;
	int i;
	struct User *user;
    struct Channel *c;
    dlink_node *ptr, *ptr2;

	count = 0;
	mem = 0;

	for (i = 0; i < 1024; i++) 
    {
        DLINK_FOREACH(ptr, user_list.head)
        {
            user = ptr->data;
			count++;
			mem += sizeof(*user);

			if (user->username)
				mem += strlen(user->username) + 1;
			
			if (user->host)
				mem += strlen(user->host) + 1;
			
			if (user->realname)
				mem += strlen(user->realname) + 1;
			
            DLINK_FOREACH(ptr2, user->chans.head)
            {
                c = ptr->data;
				mem += sizeof(*c);
            }

            DLINK_FOREACH(ptr2, user->founder_chans.head)
            {
                c = ptr->data;
				mem += sizeof(*c);
            }
		}
	}
	
	*nusers = count;
	*memuse = mem;
}

#if 0
/* Find a user by nick.  Return NULL if user could not be found. */
/* FIXME: Remove this function */

struct User *
find_user(const char *nick)
{
	struct User *user;

	user = find_client(nick);
	return(user);
}
#endif

/* Iterate over all users in the user list.  Return NULL at end of list. */

#if 0
static struct User *current;
static int next_index;

struct User *firstuser(void)
{
	next_index = 0;
	while (next_index < 1024 && current == NULL)
		current = userlist[next_index++];
	if (debug >= 3)
		slog("debug: firstuser() returning %s",
		    current ? current->nick : "NULL (end of list)");
	return(current);
}

struct User *nextuser(void)
{
	if (current)
		current = current->next;
	if (!current && next_index < 1024) {
		while (next_index < 1024 && current == NULL)
			current = userlist[next_index++];
	}
	if (debug >= 3)
		slog("debug: nextuser() returning %s",
		    current ? current->nick : "NULL (end of list)");

	return(current);
}
#endif

/* Handle a server NICK command.
 *
 * Received: NICK othy 2 1049625382 +i ~othy p362-tnt1.brs.ihug.com.au irc2.soe.uq.edu.au :othy
 *
 *	On CONNECT
 *		av[0] = nick
 *		av[1] = hop count
 *		av[2] = time of signon
 *		av[3] = usermodes
 *		av[4] = username
 *		av[5] = hostname
 *		av[6] = hosting server
 *		av[7] = realname
 *	Else:
 *		av[1] = time of change
 */

int 
do_nick(const char *source, int ac, char **av)
{
#ifdef HAVE_GETTIMEOFDAY
	// struct timeval tv;
#endif
	/* New master nick. */
	unsigned int new_nick_id = 0;
	/* Did master nick change? */
	int ni_changed = 1;
	struct User *user = NULL;

	if (EmptyString(source)) 
    {
		/* This is a new user; create a struct User structure for it. */
		if (debug)
			slog("debug: new user: %s", av[0]);

		/* First check for AKILLs. */
		if (check_akill(av[0], av[4], av[5]))
			return(0);

		/* Now check for session limits */
		if (LimitSessions && !add_session(av[0], av[5]))
			return(0);

		/* Allocate struct User structure and fill it in. */
		user = new_user(av[0]);
		user->signon = atol(av[2]);
		strscpy(user->username, av[4], USERLEN);
		strscpy(user->host, av[5], HOSTLEN);
		user->server = findserver(av[6]);
		strscpy(user->realname, av[7], REALLEN);
		user->my_signon = time(NULL);

		if (debug) 
        {
			slog("debug: got a new user, %s[%s@%s]", user->nick,
                    user->username, user->host);
			snoop(s_NickServ, "new user: %s[%s@%s]", user->nick, 
                    user->username, user->host);
		}
		display_news(user, NEWS_LOGON);
	} 
    else 
    {
		/* An old user changing nicks. */
		user = find_user(source);
		
		if (user == NULL) 
        {
			slog("user: NICK from nonexistent nick %s: %s", source, 
                    merge_args(ac, av));
			return(0);
		}
		
		if (debug)
			slog("debug: %s changes nick to %s", source, av[0]);

		/*
		 * Changing nickname case isn't a real change.  Only update
		 * my_signon if the nicks aren't the same, case-insensitively.
         * XXX ircd should do this for you.. but I guess we should play safe.
         * -stu
		 */
		if (irc_stricmp(av[0], user->nick) != 0)
			user->my_signon = time(NULL);

		new_nick_id = mysql_findnick(av[0]);
		
		if (new_nick_id)
			new_nick_id = mysql_getlink(new_nick_id);
		
		if (new_nick_id != user->nick_id) 
        {
			if (debug) 
            {
				slog("user: cancelling %s because nick "
				    "ID's differ on nick change",
				    user->nick);
				snoop(s_NickServ, "[USER] Cancelling %s "
				      "because nick ID's differ on nick "
				      "change", user->nick);
			}
			cancel_user(user);
		} 
        else
			ni_changed = 0;
		change_user_nick(user, av[0], ni_changed);
	}

	if (ni_changed) 
    {
		if (validate_user(user))
			check_memos(user);
	} 
    else 
		/* Make sure NS_ON_ACCESS flag gets set for current nick. */
		check_on_access(user);

	if (nick_identified(user)) 
    {
		if (!ni_changed && user->real_id && user->real_id != user->nick_id) 
			mysql_update_last_seen(user, 0);

		send_cmd(ServerName, "SVSMODE %s :+R", av[0]);
        SetNickReg(user);
	}

	return(1);
}

/*************************************************************************/

/* Handle a JOIN command.
 *	av[0] = channels to join
 */

void 
do_join(const char *source, int ac, char **av)
{
	time_t now;
	unsigned int channel_id;
	struct User *user;
	char *s, *t;
    struct Channel *c;
    dlink_node *ptr, *n_ptr;

	user = find_user(source);
	now = time(NULL);
	
	if (user == NULL) 
    {
		slog("user: JOIN from nonexistent user %s: %s", source, 
                merge_args(ac, av));
		return;
	}
	
	t = av[0];
	
	while (*(s = t)) 
    {
		t = s + strcspn(s, ",");
		
		if (*t)
			*t++ = '\0';
		if (debug)
			slog("debug: %s joins %s", source, s);

		if (*s == '0')
        {
            /* JOIN 0 */
            DLINK_FOREACH_SAFE(ptr, n_ptr, user->chans.head)
            {
                c = ptr->data;
				chan_deluser(user, c);
                dlinkDelete(&user->chnode, &user->chans);
				sfree(c);
			}
			continue;
		}

		/*
		 * Make sure check_kick comes before chan_adduser, so banned users
		 * don't get to see things like channel keys.
		 */
		if (check_kick(user, s))
			continue;
		
		chan_adduser(user, s);
		
		if ((channel_id = mysql_findchan(s))) 
        {
			if ((now - start_time) >= 10)
				display_entry_message(user, channel_id);
		
			check_chan_memos(user, channel_id);
		}

		c = smalloc(sizeof(*c));
        dlinkAdd(c, &user->chnode, &user->chans);
	}
}

/*************************************************************************/

/*
 * Handle an SJOIN command.
 * Basic format:
 *      av[0] = TS3 timestamp
 *      av[1] = channel
 *      av[2] = channel modes + n args(key/limit)
 *      av[3] = flags + nicklist
 */

void 
do_sjoin(const char *source, int ac, char **av)
{
	time_t now;
	int modecnt;		/* Number of modes for a user (+o and/or +v) */
	int joins;		/* Number of users that actually joined (after akick)*/
	unsigned int channel_id;
	char modebuf[4];	/* Used to hold a nice mode string */
	struct User *user;
	char *av_cmode[4];
	char *t, *nick, *channel;
    struct Channel *c;

	joins = 0;
	now = time(NULL);

	channel = av[1];
	
	if (ac < 4) 
        return;

    t = av[3];
    
	while (*(nick = t)) 
    {
		t = nick + strcspn(nick, " ");
		
		if (*t)
			*t++ = 0;

		modecnt = 0;
		modebuf[0] = '+';
		
		while (*nick == '@' || *nick == '+') 
        {
			switch (*nick) 
            {
                case '@':
                    modecnt++;
                    modebuf[modecnt] = 'o';
                    break;
                case '+':
                    modecnt++;
                    modebuf[modecnt] = 'v';
                    break;
                case '%':
                    modecnt++;
                    modebuf[modecnt] = 'h';
                    break;
            }
			nick++;
		}
		
		modebuf[modecnt + 1] = 0;
		user = find_user(nick);
		
		if (user == NULL) 
        {
			slog("channel: SJOIN to channel %s for "
			    "non-existant nick %s (%s)", channel, nick,
			    merge_args(ac - 1, av));
			continue;
		}

		if (debug)
			slog("debug: %s SJOINs %s", nick, channel);

		/*
		 * Make sure check_kick comes before chan_adduser, so
		 * banned users don't get to see things like channel
		 * keys.
		 */
		if (check_kick(user, channel))
			continue;

		chan_adduser(user, channel);
		joins++;

		if ((channel_id = mysql_findchan(channel))) 
        {
			if ((now - start_time) >= 10)
				display_entry_message(user, channel_id);

			check_chan_memos(user, channel_id);
		}

        c = findchan(channel);
        dlinkAdd(c, &user->chnode, &user->chans);

		if (modecnt > 0) 
        {
			/* Set channel modes for user. */
			if (debug) 
				slog("debug: channel modes for %s are %s", nick, modebuf);
			
			av_cmode[0] = channel;
			av_cmode[1] = modebuf;
			av_cmode[2] = nick;
			av_cmode[3] = nick;
			do_cmode(source, modecnt + 2, av_cmode);
		}
	}

	/*
	 * Did anyone actually join the channel and are there really any
	 * modes?
	 */
	if (joins && ac > 2 && av[2][1] != '\0') 
    {
		/*
		 * Set channel modes.
		 * We need to watch out for the additional params for +k and
		 * +l.  But we don't have to try and figure out which is which,
		 * because the remote server will put them in the right order
		 * for us.
		 */
		av_cmode[0] = channel;
		av_cmode[1] = av[2];
		
        av_cmode[2] = av[3];
		
		do_cmode(source, ac - 2, av_cmode);
	}
}

/*************************************************************************/

/* Handle a PART command.
 *	av[0] = channels to leave
 *	av[1] = reason (optional)
 */

void 
do_part(const char *source, int ac, char **av)
{
	struct User *user;
	char *s, *t;
    struct Channel *c;
    dlink_node *ptr;

	user = find_user(source);
	
	if (user == NULL) 
    {
		slog("user: PART from nonexistent user %s: %s", source, 
                merge_args(ac, av));
		return;
	}
	
	t = av[0];
	
	while (*(s = t)) 
    {
		t = s + strcspn(s, ",");
		
		if (*t)
			*t++ = 0;
		
		if (debug)
			slog("debug: %s leaves %s", source, s);
		
        c = NULL;
        DLINK_FOREACH(ptr, user->chans.head)
        {
            c = ptr->data;
            if(stricmp(s, c->name) == 0)
                break;
        }
		
		if (c != NULL) 
        {
			chan_deluser(user, c);
			
            dlinkDelete(&user->chnode, &user->chans);
			sfree(c);
		}
	}
}

/*************************************************************************/

/* Handle a KICK command.
 *	av[0] = channel
 *	av[1] = nick(s) being kicked
 *	av[2] = reason
 */

void 
do_kick(const char *source, int ac, char **av)
{
	struct User *user;
	char *s, *t;
    struct Channel *c;
    dlink_node *ptr;

	t = av[1];
	
	while (*(s = t)) 
    {
		t = s + strcspn(s, ",");
		
		if (*t)
			*t++ = 0;
		
		user = find_user(s);
		
		if (user == NULL) 
        {
			slog("user: KICK for nonexistent user %s on %s: %s", s,
			    av[0], merge_args(ac - 2, av + 2));
			continue;
		}
		
		if (debug)
			slog("debug: kicking %s from %s", s, av[0]);
		
        c = NULL;
        
        DLINK_FOREACH(ptr, user->chans.head)
        {
            c = ptr->data;
            if(stricmp(av[0], c->name) == 0)
                break;
        }
		if (c != NULL) 
        {
            dlinkDelete(&user->chnode, &user->chans);
			chan_deluser(user, c);
			sfree(c);
		}
	}
}

/*************************************************************************/

/* Handle a MODE command for a user.
 *	av[0] = nick to change mode for
 *	av[1] = modes
 */

void do_umode(const char *source, int ac, char **av)
{
	/* 1 if adding modes, 0 if deleting. */
	int add;
	struct User *user;
	char *s;

	add = 1;

	if (stricmp(source, av[0]) != 0) {
		slog("user: MODE %s %s from different nick %s!", av[0], av[1],
		    source);
		wallops(NULL, "%s attempted to change mode %s for %s",
		    source, av[1], av[0]);
		return;
	}
	
	user = find_user(source);
	
	if (!user) {
		slog("user: MODE %s for nonexistent nick %s: %s", av[1],
		    source, merge_args(ac, av));
		return;
	}
	
	if (debug)
		slog("debug: Changing mode for %s to %s", source, av[1]);
	
	s = av[1];
	
	while (*s) {
		switch (*s++) {
		case '+':
			add = 1;
			break;
		case '-':
			add = 0;
			break;

                case 'R':
                        add ? SetNickReg(user) : ClearNickReg(user);

                        if (add && user->real_id)
                        {
                           mysql_update_last_seen(user, NS_IDENTIFIED);
                           nickcache_update(user->real_id);
                           send_cmd(s_NickServ, "NOTICE %s :You have been automatically identified", user->nick);
                           send_cmd(ServerName, "SVSMODE %s :+R", user->nick);
                           if (strlen(get_email_from_id(user->real_id)) <= 1)
                               notice_lang(s_NickServ, user, NICK_BEG_SET_EMAIL,s_NickServ);
                           if (!(get_nick_status(user->real_id) & NS_RECOGNIZED))
                               check_memos(user);
                           if (!(get_nick_status(user->real_id) & NS_RECOGNIZED) && (get_nick_flags(user->real_id)) & NI_CLOAK) {
                               char *cloak = get_cloak_from_id(user->real_id);
                               if (cloak && *cloak != '\0') {
                                   send_cmd(ServerName, "SVSCLOAK %s :%s", user->nick, cloak);
                                   free(cloak);
                               }
                               check_autojoins(user);
                           }
                       }
                       break;
			
			
		case 'a':
			if (is_oper(user->nick) &&
			    is_services_admin(user)) {
				if (!add) {
					send_cmd(s_NickServ,
					    "SVSMODE %s :+a", av[0]);
				}
			} else {
				if (add) {
					send_cmd(s_NickServ,
					    "SVSMODE %s :-a", av[0]);
				}
			}
			break;

		case 'i':
			add ? (user->mode |= UMODE_INVISIBLE) :
			    (user->mode &= ~UMODE_INVISIBLE);
			break;
			
		case 'w':
			add ? (user->mode |= UMODE_WALLOPS) :
			    (user->mode &= ~UMODE_WALLOPS);
			break;
			
		case 'g':
			add ? (user->mode |= UMODE_GLOBAL) :
			    (user->mode &= ~UMODE_GLOBAL);
			break;
			
		case 's':
			add ? (user->mode |= UMODE_SERVNOTE) :
			    (user->mode &= ~UMODE_SERVNOTE);
			break;
			
		case 'o':
			if (add) {
                SetOper(user);
				
				if (WallOper) {
					wallops(s_OperServ,
					    "\2%s\2 is now an IRC operator.",
					    user->nick);
				}

				display_news(user, NEWS_OPER);
				opcnt++;
			} else {
				ClearOper(user);
				opcnt--;
			}
			break;
		}
	}
}

/*************************************************************************/

/* Handle a QUIT command.
 *	av[0] = reason
 */

void do_quit(const char *source, int ac, char **av)
{
	struct User *user;

	user = find_user(source);

	if (!user) {
		slog("user: QUIT from nonexistent user %s: %s", source,
		    merge_args(ac, av));
		return;
	}
	if (debug)
		slog("debug: %s quits", source);
	if (user->nick_id) {
		/* update the quit times and messages etc */
		mysql_quit_user(user, time(NULL), av[0]);
	}

	if (LimitSessions)
		del_session(user->host);
	delete_user(user);
}

/*************************************************************************/

/* Handle a KILL command.
 *	av[0] = nick being killed
 *	av[1] = reason
 */

void 
do_kill(const char *source, int ac, char **av)
{
	struct User *user;

	user = find_user(av[0]);
	
	if (!user)
		return;
	
	if (debug)
		slog("debug: %s killed", av[0]);
	
	if (user->nick_id) {
		/* Update the quit times and messages etc. */
		mysql_quit_user(user, time(NULL), av[1]);
	}
	
	if (LimitSessions)
		del_session(user->host);
	
	delete_user(user);
}

/*************************************************************************/

/*************************************************************************/

/* Is the given nick an oper? */

int is_oper(const char *nick)
{
	struct User *user;
       
	user = find_user(nick);

	return(user && IsOper(user));
}

/*************************************************************************/

/* Is the given nick on the given channel? */

int is_on_chan(const char *nick, const char *chan)
{
	struct User *u;
    struct Channel *c;
    dlink_node *ptr;

	u = find_user(nick);

	if (!u)
		return(0);
	
    DLINK_FOREACH(ptr, u->chans.head)
    {
        c = ptr->data;
		if (stricmp(c->name, chan) == 0)
			return(1);
	}
	
	return(0);
}

/*************************************************************************/

/* Is the given nick a channel operator on the given channel? */

int is_chanop(const char *nick, const char *chan)
{
	struct Channel *c;
    struct User *u;
    dlink_node *ptr;

	c = findchan(chan);

	if (!c)
		return(0);
	
    DLINK_FOREACH(ptr, c->chanops.head)
    {
        u = ptr->data;
		if (irc_stricmp(u->nick, nick) == 0)
			return(1);
	}
	
	return(0);
}

/*************************************************************************/

/* Is the given nick voiced (channel mode +v) on the given channel? */

int is_voiced(const char *nick, const char *chan)
{
	struct Channel *c;
    struct User *u;
    dlink_node *ptr;

	c = findchan(chan);

	if (!c)
		return(0);
	
    DLINK_FOREACH(ptr, c->voices.head)
    {
        u = ptr->data;
		if (irc_stricmp(u->nick, nick) == 0)
			return(1);
	}
	
	return(0);
}

/*
 * Does the user's usermask match the given mask (either nick!user@host or
 * just user@host)?
 */

int match_usermask(const char *mask, struct User *user)
{
	int result;
	char *mask2, *nick, *username, *host, *nick2, *host2;

	if (match_debug) {
		slog("match_usermask('%s', '%s!%s@%s')", mask, user->nick,
		    user->username, user->host);
	}

	mask2 = sstrdup(mask);

	if (strchr(mask2, '!')) {
		nick = strlower(strtok(mask2, "!"));
		username = strtok(NULL, "@");
	} else {
		nick = NULL;
		username = strtok(mask2, "@");
	}
	
	host = strtok(NULL, "");
	
	if (!username || !host) {
		if (match_debug) {
			slog("  mask contains missing username or missing "
			    "host.  No match");
		}
		free(mask2);
		return(0);
	}
	
	strlower(host);
	host2 = strlower(sstrdup(user->host));
	
	if (nick) {
		nick2 = strlower(sstrdup(user->nick));
		result = match_wild(nick, nick2) &&
		    match_wild(username, user->username) &&
		    (match_wild(host, host2) /*||
		    match_wild(host, user->ipstring)*/);
		free(nick2);
	} else {
		result = match_wild(username, user->username) &&
		    (match_wild(host, host2) /*||
		    match_wild(host, user->ipstring)*/);
	}
	
	free(mask2);
	free(host2);
	return(result);
}

/*************************************************************************/

/* Count the number of users who match a given mask */

int count_mask_matches(const char *mask)
{
	struct User *u;
	int i;
    dlink_node *ptr;

	i = 0;

    DLINK_FOREACH(ptr, user_list.head)
    {
        u = ptr->data;
		if (match_usermask(mask, u))
			i++;
	}

	return(i);
}

/*************************************************************************/

/*
 * Split a usermask up into its constitutent parts.  Returned strings are
 * malloc()'d, and should be free()'d when done with.  Returns "*" for
 * missing parts.
 */

void split_usermask(const char *mask, char **nick, char **user, char **host)
{
	char *mask2;

	mask2 = sstrdup(mask);

	*nick = strtok(mask2, "!");
	*user = strtok(NULL, "@");
	*host = strtok(NULL, "");
	
	/* Handle special case: mask == user@host. */
	if (*nick && !*user && strchr(*nick, '@')) {
		*nick = NULL;
		*user = strtok(mask2, "@");
		*host = strtok(NULL, "");
	}
	
	if (!*nick)
		*nick = "*";
	
	if (!*user)
		*user = "*";
	
	if (!*host)
		*host = "*";
	
	*nick = sstrdup(*nick);
	*user = sstrdup(*user);
	*host = sstrdup(*host);
	free(mask2);
}

/*************************************************************************/

/*
 * Given a user, return a mask that will most likely match any address the user
 * will have from that location.  For IP addresses, wildcards the appropriate
 * subnet mask (e.g. 35.1.1.1 -> 35.*; 128.2.1.1 -> 128.2.*); for named
 * addresses, wildcards the leftmost part of the name unless the name only
 * contains two parts.  If the username begins with a ~, delete it.  The
 * returned character string is malloc'd and should be free'd when done with.
 */

char *create_mask(struct User * u)
{
	char *mask, *s, *end;

	/*
	 * Get us a buffer the size of the username plus hostname.  The result
	 * will never be longer than this (and will often be shorter), thus we
	 * can use strcpy() and sprintf() safely.
	 */
	end = mask = smalloc(strlen(u->username) + strlen(u->host) + 2);
	end += sprintf(end, "%s@", u->username);
	
	if (strspn(u->host, "0123456789.") == strlen(u->host) &&
	    (s = strchr(u->host, '.')) && (s = strchr(s + 1, '.')) &&
	    (s = strchr(s + 1, '.')) && (!strchr(s + 1, '.'))) {
		s = sstrdup(u->host);
		*strrchr(s, '.') = 0;
		
		if (atoi(u->host) < 192)
			*strrchr(s, '.') = 0;
		
		if (atoi(u->host) < 128)
			*strrchr(s, '.') = 0;
		
		sprintf(end, "%s.*", s);
		free(s);
	} else {
		if ((s = strchr(u->host, '.')) && strchr(s + 1, '.')) {
			s = sstrdup(strchr(u->host, '.') - 1);
			*s = '*';
		} else {
			s = sstrdup(u->host);
		}
		strcpy(end, s);
		free(s);
	}
	return(mask);
}

/*************************************************************************/

/*
 * Update the quit message and last seen time for a registered &
 * recognised user who is quitting.
 */
static void mysql_quit_user(struct User *user, const time_t last_seen,
    const char *quit_msg)
{
	unsigned int fields, rows;
	MYSQL_RES *result;
	char *esc_quit_msg;

	result = smysql_bulk_query(mysqlconn, &fields, &rows,
	    "SELECT 1 FROM nick WHERE nick_id=%u && !(status & %d) && "
	    "(status & %d)", user->real_id, NS_VERBOTEN,
	    (NS_IDENTIFIED | NS_RECOGNIZED));
	mysql_free_result(result);

	if (rows) {
		esc_quit_msg = smysql_escape_string(mysqlconn,
		    *quit_msg ? quit_msg : "");
	
		result = smysql_bulk_query(mysqlconn, &fields, &rows,
		    "UPDATE nick SET last_seen=%lu, last_quit='%s', "
		    "last_quit_time=%lu WHERE nick_id=%u", last_seen,
		    esc_quit_msg, time(NULL), user->real_id);
		mysql_free_result(result);
		free(esc_quit_msg);
	}
}

/* send the channel's entry message (if any) to the user */
static void display_entry_message(struct User *user, unsigned int channel_id)
{
	MYSQL_ROW row;
	unsigned int fields, rows;
	MYSQL_RES *result;

	if (!channel_id)
		return;

	result = smysql_bulk_query(mysqlconn, &fields, &rows,
	    "SELECT entry_message, name FROM channel WHERE channel_id=%u "
	    "&& LENGTH(entry_message)", channel_id);
	
	if ((row = smysql_fetch_row(mysqlconn, result)))
		notice(s_ChanServ, user->nick, "[%s] %s", row[1], row[0]);

	mysql_free_result(result);
}

/*
 * Count the number of nicks currently in us that match a given POSIX
 * regular expression.
 */
unsigned int user_count_regex_match(regex_t *r)
{
	unsigned int cnt;
	struct User *u;
    dlink_node *ptr;

	cnt = 0;
	
    DLINK_FOREACH(ptr, user_list.head)
    {
        u = ptr->data;
		if (regexec(r, u->nick, 0, NULL, 0) == 0)
			cnt++;
	}

	return(cnt);
}

/*
 * Count the number of users in a channel.
 */
unsigned int chan_count_users(struct Channel *c)
{
	unsigned int cnt;
    dlink_node *ptr;

	if (!c)
		return(0);

    cnt = 0;
	
    DLINK_FOREACH(ptr, c->users.head)
		cnt++;

	return(cnt);
}

/*
 * Return 1 if a given user is banned on the given channel, otherwise 0.
 */
int is_banned(struct User *u, struct Channel *c)
{
	unsigned int i;

	if (!c)
		return(0);

	for (i = 0; i < c->bancount; i++) {
		if (match_usermask(c->newbans[i]->mask, u))
			return(1);
	}

	return(0);
}
